import java.util.*;
import java.io.*;

public class Tree{
    
    protected String threadId;
    protected String rootId;
    protected Hashtable<String, TreeNode> idToNode;
    protected Boolean nonProjective;
    protected int numCrossingEdges;
    
    public Tree(String tid, TreeNode root)
    {
	threadId = tid;
	rootId = root.nodeId;
	idToNode = new Hashtable<String, TreeNode>();
	idToNode.put(rootId, root);
	nonProjective = null;
	numCrossingEdges = -1;
    }

    public Tree(String tid, TreeNode root, Hashtable<String, TreeNode> idTable, boolean np, int ncross)
    {
	threadId = tid;
	rootId = root.nodeId;
	idToNode = idTable;
	nonProjective = np;
	numCrossingEdges = ncross;
    }

    public Tree(String tid, TreeNode root, boolean np, int ncross)
    {
    	threadId = tid;
    	rootId = root.nodeId;
    	idToNode = new Hashtable<String, TreeNode>();
    	idToNode.put(root.nodeId, root);
    	nonProjective = np;
    	numCrossingEdges = ncross;
    }


    public void setNonProjectivity(Boolean np, int ncross)
    {
	nonProjective = np;
	numCrossingEdges = ncross;
    }

    public void setRoot(TreeNode t)
    {
	rootId = t.nodeId;
    }

    public static Tree getEmptyTree(String tid)
    {
	//8655  SR8655  Z  NOPAR   NOHEAD  NID8655t0       false   0
	String superRootId = "SR" + tid;
	TreeNode Znode = new TreeNode(superRootId, "Z", "NOPAR", "NOHEAD", new LinkedList<String>());
	Tree emptyTree = new Tree(tid, Znode, false, 0);
	String snodeId = "NID" + tid + "t0";
	TreeNode Snode = new TreeNode(snodeId, "S", superRootId, "NOHEAD", new LinkedList<String>());
	Znode.addChild(Snode.nodeId);
	emptyTree.addNode(Snode);
	return emptyTree;
    }
	    
    public void annotateHeads()
    {
	//printTreePretty();
	//the head is the left most leaf 
	TreeNode rootNode = getRootNode();
	if(rootNode.nodeLabel.equals("Z"))
	    {
		if(rootNode.getChildrenIds().size() > 0)
		    rootNode = getNode(rootNode.getChildrenIds().get(0));
		else
		    return;
	    }
	LinkedList<TreeNode> nodesToTraverse = new LinkedList<TreeNode>(); // will have the S node and the different Xs and Cs
	nodesToTraverse.add(rootNode);
	while(nodesToTraverse.size() > 0)
	    {
		TreeNode currentlyExplored = nodesToTraverse.removeFirst();
		LinkedList<String> childrenOfCurrentlyExplored = currentlyExplored.getChildrenIds();
		if(childrenOfCurrentlyExplored.size() == 0)
		    continue;
		TreeNode leftMostChild = getNode(childrenOfCurrentlyExplored.get(0));
		LinkedList<TreeNode> pathToLeaf = new LinkedList<TreeNode>();
		while(leftMostChild.getChildrenIds().size() > 0)
		    {
			pathToLeaf.add(leftMostChild);
			for(int c = leftMostChild.getChildrenIds().size() - 1; c > 0; c--)
			    nodesToTraverse.addFirst(getNode(leftMostChild.getChildrenIds().get(c)));
			leftMostChild = getNode(leftMostChild.getChildrenIds().get(0));
		    }
		TreeNode leftMostLeaf = leftMostChild; //this will be a P node
		for(TreeNode leftPathNode: pathToLeaf)
		    leftPathNode.headId = leftMostLeaf.headId;
		currentlyExplored.headId = leftMostLeaf.headId;
		for(int c = childrenOfCurrentlyExplored.size() - 1; c > 0; c--)
		    nodesToTraverse.addFirst(getNode(childrenOfCurrentlyExplored.get(c)));
	    }
    }


    private void insertIntoOrderedList(LinkedList<Integer> list, int value)
    {
	if(value > list.get(list.size() - 1))
	    list.add(value);
	else
	    {
		list.add(value);
		for(int i = 0; i < list.size(); i++)
		    {
			for(int j = i + 1; j < list.size(); j++)
			    {
				if(list.get(i) > list.get(j))
				    {
					int temp = list.get(i);
					list.set(i, list.get(j));
					list.set(j, temp);
				    }
			    }
		    }
	    }
    }
	

    public Boolean getNonProjectivity()
    {
	return nonProjective;
    }

    public int getNumCrossingEdges()
    {
	return numCrossingEdges;
    }

    public String getThreadId()
    {
	return threadId;
    }

    public TreeNode getRootNode()
    {
	return idToNode.get(rootId);
    }

    public TreeNode getNode(String nodeId)
    {
	return idToNode.get(nodeId);
    }

    public boolean removeNode(TreeNode t)
    {
	if(idToNode.containsKey(t.nodeId))
	    {
		idToNode.remove(t.nodeId);
		return true;
	    }
	return false;
    }

    protected Hashtable<String, TreeNode> getNodeTable()
    {
	return idToNode;
    }

    public TreeNode getParent(String nodeId)
    {
	String parentId = getNode(nodeId).parentNodeId;
	if(parentId.equals("NOPAR"))
	    return null;
	else
	    return getNode(parentId);
    }

    public LinkedList<TreeNode> getChildrenNodes(String idOfParent)
    {
	LinkedList<TreeNode> ret = new LinkedList<TreeNode>();
	for(String cid: idToNode.get(idOfParent).getChildrenIds())
	    {
		TreeNode t = idToNode.get(cid);
		ret.add(t);
	    }
	return ret;
    }

    public void addNode(TreeNode t)
    {
	String parent = t.parentNodeId;
	if(!parent.equals("NOPAR") && !idToNode.get(parent).hasChildSpecified(t.nodeId))
	    idToNode.get(parent).addChild(t.nodeId);

	idToNode.put(t.nodeId, t);
    }

    public void addNodes(LinkedList<TreeNode> nodeList)
    {
	for(TreeNode t: nodeList)
	    addNode(t);
    }

    public String getBrackettedFormatString()
    {
	String bracketString = "";
	TreeNode root = idToNode.get(rootId);
	bracketString += "(" + idToNode.get(rootId).nodeLabel;
	int numBraces = 1;

	for(String childid: root.childrenNodeIds)
	    {
		String substring = getBrackettedFormatString(idToNode.get(childid));
		bracketString += " " + substring + " ";
	    }
	return bracketString.trim() + ")";
    }


    public Tree copy()
    {
	Hashtable<String, TreeNode> newIdTable = new Hashtable<String, TreeNode>();
	Enumeration et = idToNode.keys();
	while(et.hasMoreElements())
	    {
		String nodeId = (String)et.nextElement();
		TreeNode node = idToNode.get(nodeId);
		newIdTable.put(nodeId, node.copy());
	    }
	String newThreadId = getThreadId();
	TreeNode newRoot = getRootNode().copy();
	return new Tree(newThreadId, newRoot, newIdTable, nonProjective, numCrossingEdges); 
    }


    public String getBrackettedFormatString(TreeNode t)
    {
	String ret = "";
	if(t.childrenNodeIds.size() == 0)
	    ret = t.nodeId; //"(" + t.nodeLabel + " " + t.nodeId + ")";
	else
	    {
		ret = "(" + t.nodeLabel;
		for(String cid : t.childrenNodeIds)
		    ret += " " + getBrackettedFormatString(idToNode.get(cid)) + "";
		ret += ")";
	    }
	return ret;
    }

    //returns the ids of posts that are in the projection of this node
    public LinkedList<String> getProjectionOfNode(TreeNode node)
    {
	LinkedList<String> cover = new LinkedList<String>();
	if(isTerminal(node))
	    {
		cover.add(node.nodeId);
		return cover;
	    }
	//must have children
	for(TreeNode t: depthFirstTraversal(node))
	    {
		if(isTerminal(t))
		    cover.add(t.nodeId);
	    }
	return cover;
    }


    //will include the node sent in
    public LinkedList<TreeNode> depthFirstTraversal(TreeNode node)
    {
	LinkedList<TreeNode> nodesVisited = new LinkedList<TreeNode>();
	LinkedList<TreeNode> nodesToTraverse = new LinkedList<TreeNode>();
	nodesToTraverse.add(node);
	while(nodesToTraverse.size() != 0)
	    {
		TreeNode curNode = nodesToTraverse.removeFirst();
		nodesVisited.add(curNode);
		for(int c = curNode.childrenNodeIds.size() - 1; c >= 0; c--)
		    {
			TreeNode childNode = getNode(curNode.childrenNodeIds.get(c));
			nodesToTraverse.addFirst(childNode);
		    }
	    }
	return nodesVisited;
    }

    public int getNumTerminals()
    {
	int numTerminals = 0;
	Enumeration et = idToNode.keys();
	while(et.hasMoreElements())
	    {
		String id = (String) et.nextElement();
		if(isTerminal(idToNode.get(id)))
		    numTerminals++;
	    }
	return numTerminals;
    }


    public boolean isTerminal(TreeNode n)
    {
	if(n.nodeLabel.startsWith("P-") || n.nodeLabel.equals("P"))
	    return true;
	return false;
    }

    public void printTree()
    {
	System.out.println();
	LinkedList<String> nodeDetails = getStringRepOfTreeNodes();
	for(String st: nodeDetails)
	    System.out.println(st);
	System.out.println();
    }


    public LinkedList<String> getStringRepOfTreeNodes()
    {
	LinkedList<String> ret = new LinkedList<String>();
	TreeNode root = idToNode.get(rootId);
	ret.add(nodeToString(root) + "\t" + getNonProjectivity() + "\t" + getNumCrossingEdges());
	LinkedList<TreeNode> nodesToTraverse = new LinkedList<TreeNode>();
	for(String childid: root.childrenNodeIds)
	    nodesToTraverse.add(idToNode.get(childid));
	while(nodesToTraverse.size() != 0)
	    {
		TreeNode node = nodesToTraverse.removeFirst();
		ret.add(nodeToString(node));
		LinkedList<TreeNode> childrenLeftToRight = new LinkedList<TreeNode>();
		for(String childid: node.childrenNodeIds)
		    childrenLeftToRight.add(idToNode.get(childid));
		while(childrenLeftToRight.size() > 0)
		    nodesToTraverse.addFirst(childrenLeftToRight.removeLast());
	    }
	return ret;
    }

    public String nodeToString(TreeNode t)
    {
	return threadId + "\t" + t.toString();
    }
	 

    public String toString()
    {
	return idToNode.toString();
    }

    
    //non projective trees are printed with a prefix *numcrossings*
    public void printTreePretty()
    {
	System.out.println();
	//System.out.println(idToNode.toString());
	Hashtable<String, Integer> nodeIdToIndents = new Hashtable<String, Integer>();
	TreeNode root = idToNode.get(rootId);
	nodeIdToIndents.put(root.nodeId, 0);
	if(nonProjective)
	    printIndentedString(root.nodeLabel + "#" + root.nodeId + "[" + root.headId + "]" + "*" + numCrossingEdges + "*" , 0);
	else
	    printIndentedString(root.nodeLabel + "#" + root.nodeId + "[" + root.headId + "]", 0);   
	LinkedList<TreeNode> nodesToTraverse = new LinkedList<TreeNode>();
	for(String childid: root.childrenNodeIds)
	    nodesToTraverse.add(idToNode.get(childid));
	while(nodesToTraverse.size() != 0)
	    {
		TreeNode node = nodesToTraverse.removeFirst();
		int indents = nodeIdToIndents.get(node.parentNodeId) + 1;
		printIndentedString(node.nodeLabel + "#" + node.nodeId + "[" + node.headId + "]", indents);
		nodeIdToIndents.put(node.nodeId, indents);
		LinkedList<TreeNode> childrenLeftToRight = new LinkedList<TreeNode>();
		for(String childid: node.childrenNodeIds)
		    childrenLeftToRight.add(idToNode.get(childid));
		while(childrenLeftToRight.size() > 0)
		    nodesToTraverse.addFirst(childrenLeftToRight.removeLast());
	    }
	System.out.println();
    }

    public void printIndentedString(String str, int indents)
    {
	for(int i = 0; i < indents; i++)
	    System.out.print("   ");
		
	System.out.println(str);
    }

    public LinkedList<String> getPrettyTree()
    {
	LinkedList<String> ret = new LinkedList<String>();
	Hashtable<String, Integer> nodeIdToIndents = new Hashtable<String, Integer>();
	TreeNode root = idToNode.get(rootId);
	nodeIdToIndents.put(root.nodeId, 0);
	if(nonProjective)
	    ret.add(getIndentedString(root.nodeLabel + "#" + root.nodeId + "[" + root.headId + "]" + "*" + numCrossingEdges + "*", 0));
	else
	    ret.add(getIndentedString(root.nodeLabel + "#" + root.nodeId + "[" + root.headId + "]", 0));
	    
	LinkedList<TreeNode> nodesToTraverse = new LinkedList<TreeNode>();
	for(String childid: root.childrenNodeIds)
	    nodesToTraverse.add(idToNode.get(childid));
	while(nodesToTraverse.size() != 0)
	    {
		TreeNode node = nodesToTraverse.removeFirst();
		int indents = nodeIdToIndents.get(node.parentNodeId) + 1;
		ret.add(getIndentedString(node.nodeLabel + "#" + node.nodeId + "[" + node.headId + "]", indents));
		nodeIdToIndents.put(node.nodeId, indents);
		LinkedList<TreeNode> childrenLeftToRight = new LinkedList<TreeNode>();
		for(String childid: node.childrenNodeIds)
		    childrenLeftToRight.add(idToNode.get(childid));
		while(childrenLeftToRight.size() > 0)
		    nodesToTraverse.addFirst(childrenLeftToRight.removeLast());
	    }
	return ret;
    }

    public String getIndentedString(String str, int indents)
    {
	String ret = "";
	for(int i = 0; i < indents; i++)
	    ret += "   ";
	ret += str;
	return ret;
    }

    public static LinkedList<Tree> readTreesFromFile(String inputFile) throws java.io.IOException, java.io.FileNotFoundException
    {
	//System.out.println(inputFile);
	LinkedList<Tree> ret = new LinkedList<Tree>();
	BufferedReader br = new BufferedReader(new FileReader(inputFile));
	String line = "";

	Tree currentTree = null;
	while((line = br.readLine())!=null)
	    {
		String toks[] = line.trim().split("\t");
		String threadId = toks[0];
		String parent = toks[3];
		LinkedList<String> childIds = new LinkedList<String>();
		if(!toks[5].trim().equals("NOCH"))
		    {
			String chids[] = toks[5].trim().split(",");
			for(String ch: chids)
			    childIds.add(ch);
		    }
		TreeNode t = new TreeNode(toks[1], toks[2], toks[3], toks[4], childIds);
		if(currentTree == null)
		    {
			Boolean nonProj = Boolean.parseBoolean(toks[6]);
			int numCrossEdges = Integer.parseInt(toks[7]);
			currentTree = new Tree(threadId, t, nonProj, numCrossEdges);
		    }
		else if(currentTree.threadId.equals(threadId))
		    currentTree.addNode(t);
		else
		    {
			ret.add(currentTree);
			Boolean nonProj = Boolean.parseBoolean(toks[6]);
			int numCrossEdges = Integer.parseInt(toks[7]);
			currentTree = new Tree(threadId, t, nonProj, numCrossEdges);
		    }
	    }
	ret.add(currentTree);
	br.close();
	return ret;
    }
}
